---------------------------------------------------------------------------------------------------------------
-- Class: Presets ---------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------

class 'Presets'

-- Initilisation function
function Presets:__init()
  self.presets_path = nil
  self.presets_list = nil
  self.curr_preset = nil
  self.curr_preset_idx = nil  
  self.ext = "txt"
end


-- Set Path from File Selector
function Presets:PathSelector()
  local path = renoise.app():prompt_for_path("Select Presets Folder")
  return self:SetPath(path)
end


-- Set Path
function Presets:SetPath(path)
  self.presets_path = path
  if self.presets_path ~= nil then
    self:RefreshPresets()
  end
  return self.presets_path
end


-- Refresh Presets List
function Presets:RefreshPresets()

  -- Get presets from list of list of files in folder
  self.presets_list = os.filenames(self.presets_path, {"*."..self.ext})

  -- Remove file extensions from presets table
  for n = 1, #self.presets_list do
    self.presets_list[n] = string.sub(self.presets_list[n], 1, -5)
  end

  return self.presets_list
end



-- Set current preset index from file name
function Presets:SetPresetIdx(name)
  for n = 1, #self.presets_list do
    if self.presets_list[n] == name then
      self.curr_preset_idx = n
      return self.curr_preset_idx
    end
  end
end



-- Load an existing preset (from index)
function Presets:LoadPreset(preset_idx)

  -- Set path and current preset
  self.curr_preset_idx = preset_idx
  local filename = self.presets_path..self.presets_list[self.curr_preset_idx].."."..self.ext
  

  -- Read all the file contents to a string
  local f = io.open(filename, "r")
  local s = f:read("*all")
  f:close()

  -- Return as a table
  self.curr_preset = unpickle(s)
  return self.curr_preset

end



-- Save a new preset preset (by name)
function Presets:SaveNewPreset(data, name)
  
  local og_name = name
  
  -- Add file extension to name
  name = name.."."..self.ext
  
  -- Set path and current preset
  local filename = self.presets_path..name
  self.curr_preset = data
  
  -- Serialise table data and write to file
  data = pickle(data)
  local f = io.open(filename, "w")
  f:write(data)
  f:close()
  
  -- Set current preset index
  self:RefreshPresets()
  self:SetPresetIdx(og_name)
  
end



-- Overwrite an existing preset (from index)
function Presets:OverwritePreset(data, preset_idx)

  -- Set path and current preset
  self.curr_preset_idx = preset_idx
  self.curr_preset = data
  local filename = self.presets_path..self.presets_list[self.curr_preset_idx].."."..self.ext
  
  -- Serialise table data and write to file
  data = pickle(data)
  local f = io.open(filename, "w")
  f:write(data)
  f:close()

end



-- Show presets folder in OS
function Presets:ShowPresetsFolder()
  renoise.app():open_path(self.presets_path)
end







---------------------------------------------------------------------------------------------------------------
-- Pickle Library ---------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------

----------------------------------------------
-- Pickle.lua
-- A table serialization utility for lua
-- Steve Dekorte, http://www.dekorte.com, Apr 2000
-- Freeware
----------------------------------------------

function pickle(t)
  return Pickle:clone():pickle_(t)
end

Pickle = {
  clone = function (t) local nt={}; for i, v in pairs(t) do nt[i]=v end return nt end 
}

function Pickle:pickle_(root)
  if type(root) ~= "table" then 
    error("can only pickle tables, not ".. type(root).."s")
  end
  self._tableToRef = {}
  self._refToTable = {}
  local savecount = 0
  self:ref_(root)
  local s = ""

  while table.getn(self._refToTable) > savecount do
    savecount = savecount + 1
    local t = self._refToTable[savecount]
    s = s.."{\n"
    for i, v in pairs(t) do
        s = string.format("%s[%s]=%s,\n", s, self:value_(i), self:value_(v))
    end
    s = s.."},\n"
  end

  return string.format("{%s}", s)
end

function Pickle:value_(v)
  local vtype = type(v)
  if     vtype == "string" then return string.format("%q", v)
  elseif vtype == "number" then return v
  elseif vtype == "boolean" then return tostring(v)
  elseif vtype == "table" then return "{"..self:ref_(v).."}"
  else --error("pickle a "..type(v).." is not supported")
  end  
end

function Pickle:ref_(t)
  local ref = self._tableToRef[t]
  if not ref then 
    if t == self then error("can't pickle the pickle class") end
    table.insert(self._refToTable, t)
    ref = table.getn(self._refToTable)
    self._tableToRef[t] = ref
  end
  return ref
end

----------------------------------------------
-- unpickle
----------------------------------------------

function unpickle(s)
  if type(s) ~= "string" then
    error("can't unpickle a "..type(s)..", only strings")
  end
  local gentables = loadstring("return "..s)
  local tables = gentables()
  
  for tnum = 1, table.getn(tables) do
    local t = tables[tnum]
    local tcopy = {}; for i, v in pairs(t) do tcopy[i] = v end
    for i, v in pairs(tcopy) do
      local ni, nv
      if type(i) == "table" then ni = tables[i[1]] else ni = i end
      if type(v) == "table" then nv = tables[v[1]] else nv = v end
      t[i] = nil
      t[ni] = nv
    end
  end
  return tables[1]
end

