-- Offline Filter v1.3 by Aftab Hussain aka afta8 - fathand@gmail.com - 29th March 2018
-- Filter DSP code adapted from here: http://www.earlevel.com/main/2012/11/26/biquad-c-source-code/
-- Uses 'Pickle Table' Serialisation Library from here: http://lua-users.org/wiki/PickleTable
-- Some GUI icons taken from xStream by Danoise: https://renoise.com/tools/xstream
--[[============================================================================
main.lua
============================================================================]]--

-- Placeholder for the dialog
local dialog = nil

-- Placeholder to expose the ViewBuilder outside the show_dialog() function
local vb = nil

-- Reload the script whenever this file is saved. 
-- Additionally, execute the attached function.
_AUTO_RELOAD_DEBUG = function()
  
end

-- Read from the manifest.xml file.
class "RenoiseScriptingTool" (renoise.Document.DocumentNode)
  function RenoiseScriptingTool:__init()    
    renoise.Document.DocumentNode.__init(self) 
    self:add_property("Name", "Untitled Tool")
    self:add_property("Id", "Unknown Id")
  end

local manifest = RenoiseScriptingTool()
local ok,err = manifest:load_from("manifest.xml")
local tool_name = manifest:property("Name").value
local tool_id = manifest:property("Id").value


--------------------------------------------------------------------------------
-- Main functions
--------------------------------------------------------------------------------

require('Envelope')
require('Presets')

local m_max = math.max
local m_min = math.min
local m_sin = math.sin
local m_abs = math.abs
local m_e = math.exp
local m_tan = math.tan
local m_tanh = math.tanh
local m_pi = math.pi
local m_ln = math.log


-- Get the max peak value for a table of sample data
function max_peak(t)
  local max = -math.huge
  for k,v in pairs(t) do
      max = m_max(max, m_abs(v))
  end
  return max
end

-- Clamp input between two values
local function clamp(input, min_val, max_val)
  return m_min(m_max(input, min_val), max_val)
end

-- Modulo for lua arrays that start with 1
local function lmod(input, div)
  return (input-1)%div+1
end

-- Interpolate between two values a, b
function interx(a,b,p)
  return (1-p)*a+p*b
end

-- exp/log curve
function exp_log(x,s)
  return m_e(1)^(m_ln(16)*x^s)/15-1/15
end

-- Curve shaping function for GUI control
function shape(input, shape)
  return exp_log(input, interx(0.03, 5, exp_log(shape, 2)))
end


-- Return a breakpoint value from a set of points based on an input time position (0-1)
local function get_brk_val(points, pos)

  -- Set initial variables for x/y coordinates
  local x1 = 0
  local y1 = nil
  local x2 = 1
  local y2 = nil

  -- Find the two points either side of pos  
  for n = 1, #points do  
    
    -- find x1 and y1
    if pos >= points[n].time and points[n].time >= x1 then
      x1 = points[n].time
      y1 = points[n].value
    end

    -- find x2 and y2
    if pos <= points[n].time and points[n].time <= x2 then
      x2 = points[n].time
      y2 = points[n].value
    end
    
  end

  -- Check edge cases for y1 and y2
  if y1 == nil then
    y1 = y2
  end
  
  if y2 == nil then
    y2 = y1
  end  
  
  -- Calculate the relative x position (0-1) between x1 and x2
  local scale = 1/(x2-x1) 
  local xpos = scale*pos - scale/(1/x1)
  
  -- Check if xpos is a NaN value (occurs when x1 is 0), if so set xpos to 0
  if xpos ~= xpos then 
    xpos = 0
  end
  
  -- Return the breakpint value  
  return interx(y1,y2,xpos)
  
end


-- Distortion
local function dist(smp, gain, type)
  -- Apply distortion
  if type == "Clip" then
    for n = 1, #smp do
      smp[n] = clamp(smp[n]*gain, -1, 1)
    end
  elseif type == "Fold" then
    for n = 1, #smp do
      smp[n] = m_sin(smp[n]*gain)
    end
  elseif type == "Tube" then
    for n = 1, #smp do
      smp[n] = m_tanh(smp[n]*gain)
    end
  end
  return smp
end


-- Biquad Filter
local function biquad(smp, env_points, Fc1, Fc2, Q, curve, type)

  -- Init
  local K, norm, a0, a1, a2, b1, b2
  local z1, z2 = 0, 0

  -- Calculate coeficents
  local function coef(Fc, Q)
    K = m_tan(m_pi*Fc)
    norm = 1/(1+K/Q+K*K)   
    if type == "LP" then 
      a0 = K*K*norm
      a1 = 2*a0
      a2 = a0
      b1 = 2*(K*K-1)*norm
      b2 = (1-K/Q+K*K)*norm
    elseif type == "HP" then  
      a0 = 1*norm
      a1 = -2*a0
      a2 = a0
      b1 = 2*(K*K-1)*norm
      b2 = (1-K/Q+K*K)*norm
    elseif type == "BP" then 
      a0 = K/Q*norm
      a1 = 0
      a2 = -a0
      b1 = 2*(K*K-1)*norm
      b2 = (1-K/Q+K*K)*norm
    elseif type == "BR" then   
      a0 = (1+K*K)*norm
      a1 = 2*(K*K-1)*norm
      a2 = a0
      b1 = a1
      b2 = (1-K/Q+K*K)*norm
    end    
  end

  -- Biquad function
  local function process(input)
    local out = input*a0+z1
    z1 = input*a1+z2-b1*out
    z2 = input*a2-b2*out
    return out
  end

  -- Apply filter
  for n = 1, #smp do
    local env_val = get_brk_val(env_points, shape(n/#smp, curve))
    coef(interx(Fc2,Fc1,env_val), Q)        
    smp[n] = process(smp[n])
  end
  
  return smp
  
end


-- Main filter function called by tool menu/keybindings
local function filter(env_points, freq1, freq2, res, curve, type, pre_dist, pdist_type, drywet)

  -- Initialise variables
  local ins = renoise.song().selected_instrument_index  
  local smp = renoise.song().selected_sample_index
  local smp_buffer = renoise.song().instruments[ins].samples[smp].sample_buffer
  local channels = smp_buffer.number_of_channels
  local size = smp_buffer.number_of_frames 
  local sample_rate = smp_buffer.sample_rate
  local smp_data = {}
  local threshold = 1 -- Max volume threshold  
    
  -- Idiot check  
  if smp_buffer.has_sample_data == false then
    renoise.app():show_status("No sample data found")
    return
  end
        
  -- Read sample data into buffer and apply filter
  for c = 1, channels do      
    smp_data[c] = {}
    for f = 1, size do
      smp_data[c][f] = smp_buffer:sample_data(c, f)
    end  
    
    -- Apply pre distortion
    if pre_dist > 1 then
      smp_data[c] = dist(smp_data[c], pre_dist, pdist_type)
    end
    
    -- Apply filter 
    smp_data[c] = biquad(smp_data[c], env_points, freq1/sample_rate, freq2/sample_rate, res, curve, type)
  
  end
  
  -- Prevent clipping
  local max_p = {}
  for c = 1, channels do    
    max_p[c] = max_peak(smp_data[c])
  end
  max_p = max_peak(max_p)
  if max_p > threshold then
    local gr = threshold/max_p
    for c = 1, channels do
      for f = 1, size do
        smp_data[c][f] = smp_data[c][f]*gr
      end
    end
  end

  -- Do dry/wet mix
  if drywet < 1 then    
    for c = 1, channels do
      for f = 1, size do
        smp_data[c][f] = interx(smp_buffer:sample_data(c, f), smp_data[c][f], drywet)
      end
    end  
  end

  -- Write buffer to sample  
  smp_buffer:prepare_sample_data_changes()
  for c = 1, channels do      
    for f = 1, size do
      smp_buffer:set_sample_data(c, f, smp_data[c][f]) 
    end  
  end
  smp_buffer:finalize_sample_data_changes()
  
end


--------------------------------------------------------------------------------
-- GUI
--------------------------------------------------------------------------------

local function show_dialog()

  -- This block makes sure a non-modal dialog is shown once.
  -- If the dialog is already opened, it will be focused.
  if dialog and dialog.visible then
    dialog:show()
    return
  end
  
  -- The ViewBuilder is the basis
  vb = renoise.ViewBuilder()
  
  -- Declare variables
  local CONTENT_SPACING = renoise.ViewBuilder.DEFAULT_CONTROL_SPACING
  local DEFAULT_DIALOG_BUTTON_HEIGHT = renoise.ViewBuilder.DEFAULT_DIALOG_BUTTON_HEIGHT
  local DIALOG_MARGIN = renoise.ViewBuilder.DEFAULT_DIALOG_MARGIN

  local function freq(value)
    return math.floor(exp_log(value,1.5)*22050+0.5)  
  end

  -- Envelope Object
  local env = Envelope(395,140)
  
  -- Setup envelope presets
  local env_presets = Presets()
  env_presets:SetPath("envelope_presets/")

  -- Default settings
  local pre_dist = 1
  local pdist_type = "Clip"
  local distortions = {"Clip", "Fold", "Tube"}
  local filters = {"LP", "HP", "BP", "BR"}
  local type = "LP"
  local sfreq = freq(0.5)
  local efreq = freq(0.5)
  local res_mult = 1
  local res = 1*res_mult
  local res_mults = {"1","10","100","1000"}
  local curve = 0.5
  local drywet = 1

  -- Pre distortion
  local pre_dist_row = vb:row {    
    vb:text {
      width = 60,
      text = "Pre Dist"
    },
    vb:minislider {
      id = "pdist",
      width = 400,
      height = 20,
      min = 0,
      max = 100,
      value = 0,
      tooltip = "The amount of distortion applied before the sample is filtered",
      notifier = function(value)
        pre_dist = value
        vb.views["pdist_val"].value = pre_dist
      end
    },
    vb:value {
      id = "pdist_val",
      width = 35,
      value = vb.views["pdist"].value
    }            
  }
  
  -- Distortion Type
  local pdist_type_row = vb:row {    
    vb:text {
      width = 60,
      text = "Dist Type"
    },
    vb:switch {
      id = "pdist_type",
      width = 400,
      height = 20,
      value = 1,
      tooltip = "Available distortion types",
      items = distortions,
      notifier = function(value)
        pdist_type = distortions[value]
      end
    }
  }
  
  -- Filter Type
  local filters_row = vb:row {    
    vb:text {
      width = 60,
      text = "Filter Type"
    },
    vb:switch {
      id = "filters",
      width = 400,
      height = 20,
      value = 1,
      tooltip = "Available filter types: Lowpass, Highpass, Bandpass, Band Reject",
      items = filters,
      notifier = function(value)
        type = filters[value]
      end
    }
  }
  
  
  -- Filter Envelope
  local envelope_row = vb:row {
    vb:text {
      width = 60,
      text = "Filter Env"
    }
  }
  envelope_row:add_child(env.GUI)
  
  local envelope_controls_row =  vb:row {
    vb:text {
      width = 60,
      text = "Env Ctrl"
    }
  }
  envelope_controls_row:add_child(env.GUI_ADD)
  envelope_controls_row:add_child(env.GUI_REMOVE) 
  envelope_controls_row:add_child(env.GUI_INIT)
  envelope_controls_row:add_child(vb:space{width = 20})
  
  -- Envelope preset controls
  local prev_button = vb:button {
    text = "◀",
    height = 20,
    width = 20,
    tooltip = "Select previous preset",
    notifier = function()
      vb.views["preset_selector"].value = lmod(vb.views["preset_selector"].value-1, #env_presets.presets_list)     
    end
  }
  
  local next_button = vb:button {
    text = "▶",
    height = 20,
    width = 20,
    tooltip = "Select next preset",
    notifier = function()
      vb.views["preset_selector"].value = lmod(vb.views["preset_selector"].value+1, #env_presets.presets_list)
    end
  }
  
  local selector = vb:popup {
    id = "preset_selector",
    height = 20,
    width = 200,
    items = env_presets.presets_list,
    tooltip = "Select preset from list",
    notifier = function(value)
      env_presets:LoadPreset(value)
      env:SetPoints(env_presets.curr_preset)       
    end
  }
  
  local name_preset_gui = vb:column {
    margin = DIALOG_MARGIN/1.5,
    spacing = CONTENT_SPACING,
    vb:textfield {
      id = "preset_name",
      edit_mode = true,
      width = 200,
      notifier = function(value)
        env_presets:SaveNewPreset(env:GetPoints(), value)
        vb.views["preset_selector"].items = env_presets.presets_list
        vb.views["preset_selector"].value = env_presets.curr_preset_idx
        renoise.app():show_status("Preset Saved") 
      end 
    }
  }
    
  local saveas_button = vb:button {
    bitmap = "icons/save_as.bmp",
    height = 20,
    width = 20,
    tooltip = "Save as new preset",
    notifier = function()
      renoise.app():show_custom_prompt("Name preset", name_preset_gui, {"Ok"} )  
      vb.views["preset_name"].edit_mode = true    
    end
  }
  
  local save_button = vb:button {
    bitmap = "icons/save.bmp",
    height = 20,
    width = 20,
    tooltip = "Overwrite current preset",
    notifier = function()
      env_presets:OverwritePreset(env:GetPoints(), env_presets.curr_preset_idx)
      renoise.app():show_status("Preset Overwritten") 
    end
  }
  
  local show_button = vb:button {
    bitmap = "icons/reveal_folder.bmp",
    height = 20,
    width = 20,
    tooltip = "Show presets folder",
    notifier = function()
      env_presets:ShowPresetsFolder()
    end
  }
  
  envelope_controls_row:add_child(prev_button)
  envelope_controls_row:add_child(next_button)
  envelope_controls_row:add_child(selector)
  envelope_controls_row:add_child(saveas_button)
  envelope_controls_row:add_child(save_button)
  envelope_controls_row:add_child(show_button)
  
  -- Start Frequency
  local sfreq_row = vb:row {    
    vb:text {
      width = 60,
      text = "High Freq"
    },
    vb:minislider {
      id = "sfreq",
      width = 400,
      height = 20,
      min = 0,
      max = 1,
      value = 0.8,
      tooltip = "The highest frequency the filter is constrained to",
      notifier = function(value)
        sfreq = freq(value)
        vb.views["sfreq_val"].value = sfreq
      end
    },
    vb:value {
      id = "sfreq_val",
      width = 35,
      value = freq(vb.views["sfreq"].value)
    }            
  }
    
  -- End Frequency
  local efreq_row = vb:row {    
    vb:text {
      width = 60,
      text = "Low Freq"
    },
    vb:minislider {
      id = "efreq",
      width = 400,
      height = 20,
      min = 0,
      max = 1,
      value = 0.2,
      tooltip = "The lowest frequency the filter is constrained to",
      notifier = function(value)
        efreq = freq(value)
        vb.views["efreq_val"].value = efreq
      end
    },
    vb:value {
      id = "efreq_val",
      width = 35,
      value = freq(vb.views["efreq"].value)
    }            
  }
  
  -- Resonance
  local res_row = vb:row {    
    vb:text {
      width = 60,
      text = "Resonance"
    },
    vb:minislider {
      id = "res",
      width = 400,
      height = 20,
      min = 0.01,
      max = 10,
      value = 1,
      tooltip = "Resonance (or width on BP and BR)",
      notifier = function(value)
        res = value*res_mult
        vb.views["res_val"].value = res
      end
    },
    vb:value {
      id = "res_val",
      width = 35,
      value = vb.views["res"].value*res_mult
    }            
  }
  
  -- Resonance multiplier
  local res_mults_row = vb:row {    
    vb:text {
      width = 60,
      text = "Res Range"
    },
    vb:switch {
      id = "res_mult",
      width = 400,
      height = 20,
      value = 1,
      tooltip = "Multiplier for the resonance value",
      items = res_mults,
      notifier = function(value)
        res_mult = res_mults[value]
        res = vb.views["res"].value*res_mult
        vb.views["res_val"].value = res
      end
    }
  }
  
  -- Curve
  local curve_row = vb:row {    
    vb:text {
      width = 60,
      text = "Env Skew"
    },
    vb:minislider {
      id = "curve",
      width = 400,
      height = 20,
      min = 0,
      max = 1,
      value = 0.5,
      tooltip = "Warps the envelope to the left or right, a value of 0.5 has no warping",
      notifier = function(value)
        curve = value
        vb.views["curve_val"].value = curve
      end
    },
    vb:value {
      id = "curve_val",
      width = 35,
      value = vb.views["curve"].value
    }            
  }
    
  -- Dry/Wet
  local drywet_row = vb:row {    
    vb:text {
      width = 60,
      text = "Dry/Wet"
    },
    vb:minislider {
      id = "drywet",
      width = 400,
      height = 20,
      min = 0,
      max = 1,
      value = 1,
      tooltip = "The mix between the original and processed sounds (1 is 100% wet)",
      notifier = function(value)
        drywet = value
        vb.views["drywet_val"].value = drywet
      end
    },
    vb:value {
      id = "drywet_val",
      width = 35,
      value = vb.views["drywet"].value
    }            
  }
   
  -- Process button
  local process_button_row = vb:horizontal_aligner {
    mode = "center",
    
    vb:button {
      text = "Apply",
      height = 25,
      width = 150,
      
      notifier = function()
        filter(env:GetPoints(), sfreq, efreq, res, curve, type, pre_dist, pdist_type, drywet)    
      end
    }        
  } 


  -- Setting up the GUI layout
  local content = vb:column {
    margin = DIALOG_MARGIN/1.5,
    spacing = CONTENT_SPACING,
    
    vb:column{
      spacing = 2*CONTENT_SPACING,
              
      vb:column {
        spacing = CONTENT_SPACING,
        style = "group",
        margin = DIALOG_MARGIN/1.5,
        
 
        vb:column {
          id = "looppanel",
          margin = DIALOG_MARGIN/1.5,
          spacing = CONTENT_SPACING,
          style = "border",
          
          pre_dist_row,
          pdist_type_row,
          filters_row,
          envelope_row,
          envelope_controls_row,
          curve_row,
          sfreq_row,
          efreq_row,
          res_row,
          res_mults_row,
          drywet_row
          
        } 
      },
      process_button_row
    }  
  }    


  -- Key passthrough function
  local function keyhandler(dialog, key)
    return key
  end

      
  -- A custom dialog is non-modal and displays a user designed
  -- layout built with the ViewBuilder.   
  dialog = renoise.app():show_custom_dialog("Offline Filter", content, keyhandler)  
  
end

--------------------------------------------------------------------------------
-- Menu entries
--------------------------------------------------------------------------------

renoise.tool():add_menu_entry {
  name = "Main Menu:Tools:"..tool_name.."...",
  invoke = show_dialog  
}

renoise.tool():add_menu_entry {
  name = "Sample Editor:Process:"..tool_name.."...",
  invoke = show_dialog
}

--------------------------------------------------------------------------------
-- Key Binding
--------------------------------------------------------------------------------

renoise.tool():add_keybinding {
  name = "Global:Tools:" .. tool_name.."...",
  invoke = show_dialog
}

--------------------------------------------------------------------------------
-- MIDI Mapping
--------------------------------------------------------------------------------

--[[
renoise.tool():add_midi_mapping {
  name = tool_id..":Show Dialog...",
  invoke = show_dialog
}
--]]
